from sqlalchemy import and_, distinct, func, select

from orm.author import Author
from orm.shout import ShoutTopic
from orm.topic import Topic, TopicFollower
from resolvers.stat import get_with_stat
from services.auth import login_required
from services.db import local_session
from services.schema import mutation, query
from services.logger import root_logger as logger


@query.field('get_topics_all')
def get_topics_all(_, _info):
    q = select(Topic)
    q = q.group_by(Topic.id)
    topics = get_with_stat(q, Author, TopicFollower)
    return topics


@query.field('get_topics_by_community')
async def get_topics_by_community(_, _info, community_id: int):
    q = select(Topic).where(Topic.community == community_id)
    q = q.group_by(Topic.id)
    topics = await get_with_stat(q, Author, TopicFollower)
    return topics


@query.field('get_topics_by_author')
async def get_topics_by_author(_, _info, author_id=None, slug='', user=''):
    q = select(Topic)
    if author_id:
        q = q.join(Author).where(Author.id == author_id)
    elif slug:
        q = q.join(Author).where(Author.slug == slug)
    elif user:
        q = q.join(Author).where(Author.user == user)

    q = q.group_by(Topic.id)
    topics = await get_with_stat(q, Author, TopicFollower)
    return topics


@query.field('get_topic')
async def get_topic(_, _info, slug):
    q = select(Topic).filter(Topic.slug == slug)
    q = q.group_by(Topic.id)
    topics = await get_with_stat(q, Author, TopicFollower)
    if topics:
        return topics[0]


@mutation.field('create_topic')
@login_required
async def create_topic(_, _info, inp):
    with local_session() as session:
        # TODO: check user permissions to create topic for exact community
        # and actor is permitted to craete it
        new_topic = Topic(**inp)
        session.add(new_topic)
        session.commit()

        return {'topic': new_topic}


@mutation.field('update_topic')
@login_required
async def update_topic(_, _info, inp):
    slug = inp['slug']
    with local_session() as session:
        topic = session.query(Topic).filter(Topic.slug == slug).first()
        if not topic:
            return {'error': 'topic not found'}
        else:
            Topic.update(topic, inp)
            session.add(topic)
            session.commit()

            return {'topic': topic}


@mutation.field('delete_topic')
@login_required
async def delete_topic(_, info, slug: str):
    user_id = info.context['user_id']
    with local_session() as session:
        t: Topic = session.query(Topic).filter(Topic.slug == slug).first()
        if not t:
            return {'error': 'invalid topic slug'}
        author = session.query(Author).filter(Author.user == user_id).first()
        if author:
            if t.created_by != author.id:
                return {'error': 'access denied'}

            session.delete(t)
            session.commit()

            return {}
    return {'error': 'access denied'}


def topic_follow(follower_id, slug):
    try:
        with local_session() as session:
            topic = session.query(Topic).where(Topic.slug == slug).one()
            _following = TopicFollower(topic=topic.id, follower=follower_id)
            return True
    except Exception as _exc:
        return False


def topic_unfollow(follower_id, slug):
    try:
        with local_session() as session:
            sub = (
                session.query(TopicFollower)
                .join(Topic)
                .filter(and_(TopicFollower.follower == follower_id, Topic.slug == slug))
                .first()
            )
            if sub:
                session.delete(sub)
                session.commit()
                return True
    except Exception as ex:
        logger.debug(ex)
    return False


@query.field('get_topics_random')
async def get_topics_random(_, info, amount=12):
    q = select(Topic)
    q = q.join(ShoutTopic)
    q = q.group_by(Topic.id)
    q = q.having(func.count(distinct(ShoutTopic.shout)) > 2)
    q = q.order_by(func.random()).limit(amount)

    topics = []
    with local_session() as session:
        for [topic] in session.execute(q):
            topics.append(topic)

    return topics


def get_random_topic():
    q = select(Topic)
    q = q.join(ShoutTopic)
    q = q.group_by(Topic.id)
    q = q.having(func.count(distinct(ShoutTopic.shout)) > 10)
    q = q.order_by(func.random()).limit(1)

    with local_session() as session:
        r = session.execute(q).first()
        if r:
            [topic] = r
            return topic
    return
